/*
Copyright (c) 2001-2005 by Juliusz Chroboczek

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>
#include <stdarg.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <locale.h>

#ifdef SVR4
#include <stropts.h>
#endif

#ifdef __GLIBC__
#define HAVE_NL_LANGINFO
#include <langinfo.h>
#endif

#include "pluto.h"
#include "convert.h"
#include "sys.h"

static char *child_argv0 = NULL;
int verbose = 0;
int converter = 0;
int exitOnChild = 0;
int timeout = 20;
int guess = 1;

void *inputState, *outputState;

volatile int sigwinch_queued = 0;
volatile int sigchld_queued = 0;

static int convert(int, int);
static int condom(int, char**);

static void
FatalError(char *f, ...)
{
    va_list args;
    va_start(args, f);
    vfprintf(stderr, f, args);
    va_end(args);
    exit(1);
}


static void
help(void)
{
    fprintf(stderr, "pluto [options] program [args]\n");
}
            

static int
parseOptions(int argc, char **argv)
{
    int i = 1;
    while(i < argc) {
        if(argv[i][0] != '-' && argv[i][0] != '+') {
            break;
        } else if(!strcmp(argv[i], "--")) {
            i++;
            break;
        } else if(!strcmp(argv[i], "-v")) {
            verbose++;
            i++;
        } else if(!strcmp(argv[i], "-h")) {
            help();
            exit(0);
        } else if(!strcmp(argv[i], "-x")) {
            exitOnChild = 1;
            i++;
        } else if(!strcmp(argv[i], "-e")) {
            if(i + 1 >= argc)
                FatalError("-e requires an argument\n");
            localeEncoding = parseLocaleEncoding(argv[i]);
            i += 2;
        } else if(!strcmp(argv[i], "-p")) {
            guess = 0;
            i++;
        } else if(!strcmp(argv[i], "-t")) {
            if(i + 1 >= argc)
                FatalError("-t requires an argument\n");
            timeout = atoi(argv[i + 1]);
            i += 2;
        } else if(!strcmp(argv[i], "-c")) {
            converter = 1;
            i++;
        }
    }
    return i;
}

static int
parseArgs(int argc, char **argv, char *argv0,
          char **path_return, char ***argv_return)
{
    char *path = NULL;
    char **child_argv = NULL;

    if(argc <= 0) {
        char *shell;
        shell = getenv("SHELL");
        if(shell) {
            path = malloc(strlen(shell) + 1);
            if(!path)
                goto bail;
            strcpy(path, shell);
        } else {
            path = malloc(strlen("/bin/sh") + 1);
            if(!path)
                goto bail;
            strcpy(path, "/bin/sh");
        }
        child_argv = malloc(2 * sizeof(char*));
        if(!child_argv)
            goto bail;
        if(argv0)
            child_argv[0] = argv0;
        else
            child_argv[0] = my_basename(path);
        child_argv[1] = NULL;
    } else {
        path = malloc(strlen(argv[0]) + 1);
        if(!path)
            goto bail;
        strcpy(path, argv[0]);
        child_argv = malloc((argc + 1) * sizeof(char*));
        if(!child_argv) {
            goto bail;
        }
        if(child_argv0)
            child_argv[0] = argv0;
        else
            child_argv[0] = my_basename(argv[0]);
        memcpy(child_argv + 1, argv + 1, (argc - 1) * sizeof(char*));
        child_argv[argc] = NULL;
    }

    *path_return = path;
    *argv_return = child_argv;
    return 0;

  bail:
    if(path)
        free(path);
    if(argv)
        free(argv);
    return -1;
}

void
intuitEncoding()
{
    char *l, *locale_name, *encoding;
    int n;
    
    l = setlocale(LC_ALL, "");
    if(!l)
        fprintf(stderr, "Warning: couldn't set locale.\n");

 again:
    if(l) {
        locale_name = setlocale(LC_CTYPE, NULL);
    } else {
        locale_name = getenv("LC_ALL");
        if(locale_name == NULL) {
            locale_name = getenv("LC_CTYPE");
            if(locale_name == NULL) {
                locale_name = getenv("LANG");
            }
        }
    }

    if(locale_name == NULL)
        locale_name = "en_US";

    n = strlen(locale_name);
    if(n >= 4 && strcmp(locale_name + n - 4, ".UTF8") == 0) {
        locale_name[n - 4] = '\0';
        l = setlocale(LC_CTYPE, locale_name);
        if(l)
            goto again;
    }

    encoding = NULL;
#ifdef HAVE_NL_LANGINFO
    if(l)
        encoding = nl_langinfo(CODESET);
#endif


    if(locale_name != NULL && encoding == NULL) {
        n = strlen(locale_name);
        if(strcmp(locale_name, "en_US") == 0)
            encoding = "ISO 8859-1";
        else if(strcmp(locale_name, "fr_FR") == 0)
            encoding = "ISO 8859-1";
        else if(strcmp(locale_name, "pl_PL") == 0)
            encoding = "ISO 8859-2";
        else if(strcmp(locale_name, "eo_EO") == 0)
            encoding = "ISO 8859-3";
        else if(n >= 5 && strcmp(locale_name + n - 5, "@euro") == 0)
            encoding = "ISO 8859-15";
    }

    if(encoding == NULL) {
        encoding = "ISO 8859-1";
    }

    localeEncoding = parseLocaleEncoding(encoding);
}

int
main(int argc, char **argv)
{
    int i;

    intuitEncoding();
    
    i = parseOptions(argc, argv);
    if(i < 0)
        FatalError("Couldn't parse options\n");

    if(converter)
        return convert(0, 1);
    else
        return condom(argc - i, argv + i);
}

static int
convert(int ifd, int ofd)
{
    int rc, i;
    unsigned char buf[BUFFER_SIZE];

    rc = droppriv();
    if(rc < 0) {
        perror("Couldn't drop priviledges");
        exit(1);
    }

    i = 0;
    while(1) {
        rc = read(ifd, buf + i, BUFFER_SIZE - i);
        if(rc < 0 && errno == EINTR)
            continue;
        if(rc <= 0) {
            if(rc < 0) {
                perror("Read error");
                exit(1);
            }
            break;
        }
        i += rc;
        rc = copyOut(ofd, buf, i, 0);
        i = rc;
    }
    if(i > 0)
        copyOut(ofd, buf, i, 1);
    return 0;
}
        

static int
condom(int argc, char **argv)
{
    int pty;
    int pid;
    char *line;
    char *path;
    char **child_argv;
    int rc;

    rc = parseArgs(argc, argv, child_argv0,
                   &path, &child_argv);
    if(rc < 0)
        FatalError("Couldn't parse arguments\n");

    rc = allocatePty(&pty, &line);
    if(rc < 0) {
        perror("Couldn't allocate pty");
        exit(1);
    }

    rc = droppriv();
    if(rc < 0) {
        perror("Couldn't drop priviledges");
        exit(1);
    }

    pid = fork();
    if(pid < 0) {
        perror("Couldn't fork");
        exit(1);
    }

    if(pid == 0) {
        close(pty);
        child(line, path, child_argv);
    } else {
        free(child_argv);
        free(path);
        free(line);
        parent(pid, pty);
    }

    return 0;
}

void
child(char *line, char *path, char **argv)
{
    int tty;
    int pgrp;

    close(0);
    close(1);
    close(2);
    pgrp = setsid();
    if(pgrp < 0) {
        kill(getppid(), SIGABRT);
        exit(1);
    }

    tty = openTty(line);
    if(tty < 0) {
        kill(getppid(), SIGABRT);
        exit(1);
    }
    
    if(tty != 0)
        dup2(tty, 0);
    if(tty != 1)
        dup2(tty, 1);
    if(tty != 2)
        dup2(tty, 2);

    if(tty > 2)
        close(tty);
    
    execvp(path, argv);
    perror("Couldn't exec");
    exit(1);
}

static void
sigwinchHandler(int sig) {
    sigwinch_queued = 1;
}

static void
sigchldHandler(int sig)
{
    sigchld_queued = 1;
}

void
parent(int pid, int pty)
{
    unsigned char ibuf[BUFFER_SIZE], obuf[BUFFER_SIZE];
    int i, n;
    int val;
    int rc;

#ifdef SIGWINCH
    installHandler(SIGWINCH, sigwinchHandler);
#endif
    installHandler(SIGCHLD, sigchldHandler);

    rc = copyTermios(0, pty);
    if(rc < 0)
        FatalError("Couldn't copy terminal settings\n");

    rc = setRawTermios();
    if(rc < 0)
        FatalError("Couldn't set terminal to raw\n");

    val = fcntl(0, F_GETFL, 0);
    if(val >= 0) {
        fcntl(0, F_SETFL, val | O_NONBLOCK);
    }
    val = fcntl(pty, F_GETFL, 0);
    if(val >= 0) {
        fcntl(pty, F_SETFL, val | O_NONBLOCK);
    }

    setWindowSize(0, pty);

    n = 0;
    for(;;) {
        rc = waitForInput(0, pty, n > 0 ? timeout : -1);

        if(sigwinch_queued) {
            sigwinch_queued = 0;
            setWindowSize(0, pty);
        }

        if(sigchld_queued && exitOnChild)
            break;

        if(rc < 0) {
            if(errno == EINTR)
                continue;
            else
                break;
        }

        if((rc & 2) || (rc == 0 && n > 0)) {
            i = read(pty, obuf + n, BUFFER_SIZE - n);
            if((i < 0) && ((errno == EAGAIN) || (errno == EINTR)))
                i = 0;
            if(i < 0)
                break;
            n += i;
            if(n > 0)
                n = copyOut(0, obuf, i, (rc == 0));
        }
        if(rc & 1) {
            i = read(0, ibuf, BUFFER_SIZE);
            if((i == 0) || 
               ((i < 0) && (errno != EAGAIN) && (errno != EINTR)))
                break;
            if(i > 0)
                copyIn(pty, ibuf, i);
        }
    }
    restoreTermios();
}
